"""
Tests for device string patching (cuda -> musa translation).
"""

import pytest


class TestTensorDevicePatching:
    """Test tensor device string patching."""

    def test_tensor_cuda_method(self):
        """Test .cuda() method on tensors."""
        import torch

        import torchada

        if torch.cuda.is_available():
            x = torch.randn(2, 2)
            y = x.cuda()
            assert y.device.type in ("cuda", "musa")
            if torchada.is_musa_platform():
                assert y.device.type == "musa"

    def test_tensor_to_cuda_string(self):
        """Test .to('cuda') on tensors."""
        import torch

        import torchada

        if torch.cuda.is_available():
            x = torch.randn(2, 2)
            y = x.to("cuda")
            assert y.device.type in ("cuda", "musa")
            if torchada.is_musa_platform():
                assert y.device.type == "musa"

    def test_tensor_to_cuda_device_index(self):
        """Test .to('cuda:0') on tensors."""
        import torch

        import torchada

        if torch.cuda.is_available():
            x = torch.randn(2, 2)
            y = x.to("cuda:0")
            assert y.device.type in ("cuda", "musa")
            if torchada.is_musa_platform():
                assert y.device.type == "musa"
                assert y.device.index == 0

    def test_tensor_to_torch_device(self):
        """Test .to(torch.device('cuda')) on tensors."""
        import torch

        import torchada

        if torch.cuda.is_available():
            x = torch.randn(2, 2)
            y = x.to(torch.device("cuda"))
            assert y.device.type in ("cuda", "musa")


class TestModuleDevicePatching:
    """Test nn.Module device patching."""

    def test_module_cuda_method(self):
        """Test .cuda() method on modules."""
        import torch
        import torch.nn as nn

        import torchada

        if torch.cuda.is_available():
            model = nn.Linear(10, 5)
            model = model.cuda()
            param_device = next(model.parameters()).device
            assert param_device.type in ("cuda", "musa")
            if torchada.is_musa_platform():
                assert param_device.type == "musa"

    def test_module_to_cuda(self):
        """Test .to('cuda') method on modules."""
        import torch
        import torch.nn as nn

        import torchada

        if torch.cuda.is_available():
            model = nn.Linear(10, 5)
            model = model.to("cuda")
            param_device = next(model.parameters()).device
            assert param_device.type in ("cuda", "musa")


class TestFactoryFunctions:
    """Test tensor factory function patching."""

    def test_torch_empty_device_cuda(self):
        """Test torch.empty(device='cuda')."""
        import torch

        import torchada

        if torch.cuda.is_available():
            x = torch.empty(2, 2, device="cuda")
            assert x.device.type in ("cuda", "musa")
            if torchada.is_musa_platform():
                assert x.device.type == "musa"

    def test_torch_zeros_device_cuda(self):
        """Test torch.zeros(device='cuda')."""
        import torch

        import torchada

        if torch.cuda.is_available():
            try:
                x = torch.zeros(2, 2, device="cuda")
                assert x.device.type in ("cuda", "musa")
            except RuntimeError as e:
                # MUSA driver issue, not torchada
                if "MUDNN" in str(e) or "invalid device function" in str(e):
                    pytest.skip(f"MUSA driver issue: {e}")
                raise

    def test_torch_ones_device_cuda(self):
        """Test torch.ones(device='cuda')."""
        import torch

        import torchada

        if torch.cuda.is_available():
            try:
                x = torch.ones(2, 2, device="cuda")
                assert x.device.type in ("cuda", "musa")
            except RuntimeError as e:
                if "MUDNN" in str(e) or "invalid device function" in str(e):
                    pytest.skip(f"MUSA driver issue: {e}")
                raise

    def test_torch_randn_device_cuda(self):
        """Test torch.randn(device='cuda')."""
        import torch

        import torchada

        if torch.cuda.is_available():
            try:
                x = torch.randn(2, 2, device="cuda")
                assert x.device.type in ("cuda", "musa")
            except RuntimeError as e:
                if "MUSA" in str(e) or "invalid device function" in str(e):
                    pytest.skip(f"MUSA driver issue: {e}")
                raise

    def test_torch_rand_device_cuda(self):
        """Test torch.rand(device='cuda')."""
        import torch

        import torchada

        if torch.cuda.is_available():
            try:
                x = torch.rand(2, 2, device="cuda")
                assert x.device.type in ("cuda", "musa")
            except RuntimeError as e:
                if "MUSA" in str(e) or "invalid device function" in str(e):
                    pytest.skip(f"MUSA driver issue: {e}")
                raise

    def test_torch_full_device_cuda(self):
        """Test torch.full(device='cuda')."""
        import torch

        import torchada

        if torch.cuda.is_available():
            try:
                x = torch.full((2, 2), 3.14, device="cuda")
                assert x.device.type in ("cuda", "musa")
            except RuntimeError as e:
                if "MUDNN" in str(e) or "invalid device function" in str(e):
                    pytest.skip(f"MUSA driver issue: {e}")
                raise

    def test_torch_arange_device_cuda(self):
        """Test torch.arange(device='cuda')."""
        import torch

        import torchada

        if torch.cuda.is_available():
            try:
                x = torch.arange(10, device="cuda")
                assert x.device.type in ("cuda", "musa")
            except RuntimeError as e:
                if "MUSA" in str(e) or "invalid device function" in str(e):
                    pytest.skip(f"MUSA driver issue: {e}")
                raise

    def test_torch_linspace_device_cuda(self):
        """Test torch.linspace(device='cuda')."""
        import torch

        import torchada

        if torch.cuda.is_available():
            try:
                x = torch.linspace(0, 1, 10, device="cuda")
                assert x.device.type in ("cuda", "musa")
            except RuntimeError as e:
                if "MUSA" in str(e) or "invalid device function" in str(e):
                    pytest.skip(f"MUSA driver issue: {e}")
                raise


class TestTranslateDeviceFunction:
    """Unit tests for the _translate_device function."""

    def test_translate_device_none(self):
        """Test that None is passed through unchanged."""
        from torchada._patch import _translate_device

        assert _translate_device(None) is None

    def test_translate_device_cuda_string(self):
        """Test 'cuda' string translation."""
        import torchada
        from torchada._patch import _translate_device

        result = _translate_device("cuda")
        if torchada.is_musa_platform():
            assert result == "musa"
        else:
            assert result == "cuda"

    def test_translate_device_cuda_with_index(self):
        """Test 'cuda:0' and 'cuda:1' string translation."""
        import torchada
        from torchada._patch import _translate_device

        result0 = _translate_device("cuda:0")
        result1 = _translate_device("cuda:1")

        if torchada.is_musa_platform():
            assert result0 == "musa:0"
            assert result1 == "musa:1"
        else:
            assert result0 == "cuda:0"
            assert result1 == "cuda:1"

    def test_translate_device_cpu_unchanged(self):
        """Test that 'cpu' is unchanged."""
        from torchada._patch import _translate_device

        assert _translate_device("cpu") == "cpu"

    def test_translate_device_musa_unchanged(self):
        """Test that 'musa' is unchanged."""
        from torchada._patch import _translate_device

        assert _translate_device("musa") == "musa"
        assert _translate_device("musa:0") == "musa:0"

    def test_translate_device_integer_unchanged(self):
        """Test that integer device IDs are unchanged."""
        from torchada._patch import _translate_device

        assert _translate_device(0) == 0
        assert _translate_device(1) == 1

    def test_translate_device_torch_device_cuda(self):
        """Test torch.device('cuda') translation."""
        import torch

        import torchada

        # Get original torch.device class
        from torchada._patch import _original_torch_device, _translate_device

        if _original_torch_device is None:
            # Not patched yet, use current torch.device
            device_cls = torch.device
        else:
            device_cls = _original_torch_device

        cuda_device = device_cls("cuda")
        result = _translate_device(cuda_device)

        if torchada.is_musa_platform():
            assert result.type == "musa"
        else:
            assert result.type == "cuda"

    def test_translate_device_torch_device_cuda_with_index(self):
        """Test torch.device('cuda', 0) translation."""
        import torch

        import torchada
        from torchada._patch import _original_torch_device, _translate_device

        if _original_torch_device is None:
            device_cls = torch.device
        else:
            device_cls = _original_torch_device

        cuda_device = device_cls("cuda", 0)
        result = _translate_device(cuda_device)

        if torchada.is_musa_platform():
            assert result.type == "musa"
            assert result.index == 0
        else:
            assert result.type == "cuda"
            assert result.index == 0

    def test_translate_device_torch_device_cpu_unchanged(self):
        """Test torch.device('cpu') is unchanged."""
        import torch

        from torchada._patch import _translate_device

        cpu_device = torch.device("cpu")
        result = _translate_device(cpu_device)
        assert result.type == "cpu"


class TestDeviceIndexVariants:
    """Test various cuda:N device index formats."""

    def test_cuda_colon_zero(self):
        """Test cuda:0 works."""
        import torch

        import torchada

        if torch.cuda.is_available():
            x = torch.randn(2, 2)
            y = x.to("cuda:0")
            assert y.device.type in ("cuda", "musa")
            assert y.device.index == 0

    def test_factory_with_cuda_index(self):
        """Test torch.empty(device='cuda:0')."""
        import torch

        import torchada

        if torch.cuda.is_available():
            x = torch.empty(2, 2, device="cuda:0")
            assert x.device.type in ("cuda", "musa")
            assert x.device.index == 0

    def test_torch_device_cuda_zero(self):
        """Test torch.device('cuda:0')."""
        import torch

        import torchada

        device = torch.device("cuda:0")
        if torchada.is_musa_platform():
            assert device.type == "musa"
            assert device.index == 0
        else:
            assert device.type == "cuda"
            assert device.index == 0

    def test_torch_device_with_index_arg(self):
        """Test torch.device('cuda', 0)."""
        import torch

        import torchada

        device = torch.device("cuda", 0)
        if torchada.is_musa_platform():
            assert device.type == "musa"
            assert device.index == 0
        else:
            assert device.type == "cuda"
            assert device.index == 0


class TestDeviceContextManager:
    """Test device context manager (with torch.device(...):) works."""

    def test_device_context_with_musa(self):
        """Test that device context manager works with musa device."""
        import torch

        import torchada

        if not torchada.is_musa_platform():
            pytest.skip("MUSA platform required")

        device = torch.device("musa:0")
        with device:
            t = torch.empty(10)
            assert t.device.type == "musa"
            assert t.device.index == 0

    def test_device_context_with_cuda_translated(self):
        """Test that device context manager works with cuda device (translated to musa)."""
        import torch

        import torchada

        if not torchada.is_musa_platform():
            pytest.skip("MUSA platform required")

        # torch.device("cuda:0") gets translated to musa:0
        device = torch.device("cuda:0")
        assert device.type == "musa"  # Verify translation happened

        with device:
            t = torch.empty(10)
            assert t.device.type == "musa"
            assert t.device.index == 0

    def test_original_functions_in_device_constructors(self):
        """Test that original functions are added to _device_constructors."""
        import torch

        import torchada

        if not torchada.is_musa_platform():
            pytest.skip("MUSA platform required")

        try:
            from torch.utils._device import _device_constructors

            constructors = _device_constructors()

            # Check that original (unwrapped) functions are in constructors
            if hasattr(torch.empty, "__wrapped__"):
                assert torch.empty.__wrapped__ in constructors
            if hasattr(torch.zeros, "__wrapped__"):
                assert torch.zeros.__wrapped__ in constructors
            if hasattr(torch.ones, "__wrapped__"):
                assert torch.ones.__wrapped__ in constructors
        except ImportError:
            pytest.skip("torch.utils._device not available")



"""
Tests for building CUDA extensions with torchada.

These tests verify that CUDAExtension and BuildExtension work correctly
on MUSA platforms, including source code porting.

The key point is that after importing torchada, the standard torch imports
should work transparently:
    from torch.utils.cpp_extension import CUDAExtension, BuildExtension
"""

import os
import shutil
import subprocess
import sys
import tempfile

import pytest

# Import torchada first to apply patches
import torchada  # noqa: F401

# Get the path to the test CUDA source file
CSRC_DIR = os.path.join(os.path.dirname(__file__), "csrc")
VECTOR_ADD_CU = os.path.join(CSRC_DIR, "vector_add.cu")


class TestExtensionBuildSetup:
    """Test extension build setup and configuration."""

    def test_vector_add_cu_exists(self):
        """Test that vector_add.cu test file exists."""
        assert os.path.exists(VECTOR_ADD_CU), f"Test file not found: {VECTOR_ADD_CU}"

    def test_can_create_setup_py(self):
        """Test that we can create a setup.py for the extension."""
        # Use standard torch imports - torchada patches make them work on MUSA
        from torch.utils.cpp_extension import BuildExtension, CUDAExtension

        # Create a temporary directory for the test
        with tempfile.TemporaryDirectory() as tmpdir:
            # Copy the source file
            shutil.copy(VECTOR_ADD_CU, tmpdir)

            # Create setup.py content - uses standard torch imports
            setup_content = f"""
import torchada  # noqa: F401 - Apply MUSA patches
from setuptools import setup
from torch.utils.cpp_extension import CUDAExtension, BuildExtension

setup(
    name="test_vector_add",
    ext_modules=[
        CUDAExtension(
            name="test_vector_add",
            sources=["vector_add.cu"],
        )
    ],
    cmdclass={{"build_ext": BuildExtension}},
)
"""
            setup_path = os.path.join(tmpdir, "setup.py")
            with open(setup_path, "w") as f:
                f.write(setup_content)

            assert os.path.exists(setup_path)

            # Verify the setup.py is valid Python
            result = subprocess.run(
                [sys.executable, "-m", "py_compile", setup_path],
                capture_output=True,
                text=True,
            )
            assert result.returncode == 0, f"Setup.py syntax error: {result.stderr}"


def _is_gpu_available():
    """Check if CUDA or MUSA GPU is available."""
    import torch

    if torch.cuda.is_available():
        return True
    if hasattr(torch, "musa") and torch.musa.is_available():
        return True
    return False


@pytest.mark.skipif(
    not os.environ.get("TORCHADA_TEST_BUILD", "0") == "1",
    reason="Extension build tests are slow; set TORCHADA_TEST_BUILD=1 to run",
)
class TestExtensionBuild:
    """Test actual extension building (slow, opt-in)."""

    def test_build_vector_add_extension(self):
        """Test building the vector_add extension."""
        if not _is_gpu_available():
            pytest.skip("CUDA/MUSA not available")

        with tempfile.TemporaryDirectory() as tmpdir:
            # Copy the source file
            shutil.copy(VECTOR_ADD_CU, tmpdir)

            # Create setup.py using standard torch imports
            setup_content = """
import torchada  # noqa: F401 - Apply MUSA patches
from setuptools import setup
from torch.utils.cpp_extension import CUDAExtension, BuildExtension

setup(
    name="test_vector_add",
    ext_modules=[
        CUDAExtension(
            name="test_vector_add",
            sources=["vector_add.cu"],
        )
    ],
    cmdclass={"build_ext": BuildExtension},
)
"""
            setup_path = os.path.join(tmpdir, "setup.py")
            with open(setup_path, "w") as f:
                f.write(setup_content)

            # Build the extension
            result = subprocess.run(
                [sys.executable, "setup.py", "build_ext", "--inplace"],
                cwd=tmpdir,
                capture_output=True,
                text=True,
                timeout=300,  # 5 minute timeout
            )

            if result.returncode != 0:
                print("STDOUT:", result.stdout)
                print("STDERR:", result.stderr)

            assert result.returncode == 0, f"Build failed: {result.stderr}"

            # Check that the extension was built
            ext_files = [f for f in os.listdir(tmpdir) if f.endswith(".so") or f.endswith(".pyd")]
            assert len(ext_files) > 0, "No extension file was built"

    def test_run_vector_add_extension(self):
        """Test running the vector_add extension after building."""
        import torch

        if not _is_gpu_available():
            pytest.skip("CUDA/MUSA not available")

        with tempfile.TemporaryDirectory() as tmpdir:
            # Copy the source file
            shutil.copy(VECTOR_ADD_CU, tmpdir)

            # Create setup.py using standard torch imports
            setup_content = """
import torchada  # noqa: F401 - Apply MUSA patches
from setuptools import setup
from torch.utils.cpp_extension import CUDAExtension, BuildExtension

setup(
    name="test_vector_add",
    ext_modules=[
        CUDAExtension(
            name="test_vector_add",
            sources=["vector_add.cu"],
        )
    ],
    cmdclass={"build_ext": BuildExtension},
)
"""
            setup_path = os.path.join(tmpdir, "setup.py")
            with open(setup_path, "w") as f:
                f.write(setup_content)

            # Build the extension
            result = subprocess.run(
                [sys.executable, "setup.py", "build_ext", "--inplace"],
                cwd=tmpdir,
                capture_output=True,
                text=True,
                timeout=300,
            )
            assert result.returncode == 0, f"Build failed: {result.stderr}"

            # Add tmpdir to Python path and import the extension
            sys.path.insert(0, tmpdir)
            try:
                import test_vector_add

                try:
                    # Create test tensors
                    a = torch.randn(1000, device="cuda")
                    b = torch.randn(1000, device="cuda")

                    # Run vector add
                    c = test_vector_add.vector_add(a, b)

                    # Verify result
                    expected = a + b
                    assert torch.allclose(c, expected), "Vector add result incorrect"
                except RuntimeError as e:
                    # Skip if GPU is not working or kernel was compiled for different architecture
                    if "invalid device function" in str(e):
                        pytest.skip(
                            "GPU not available or kernel compiled for different architecture"
                        )
                    raise
            finally:
                sys.path.remove(tmpdir)



"""
CUDA to MUSA mapping rules for source code porting.

This module contains the comprehensive mapping dictionary for converting
CUDA-specific symbols to their MUSA equivalents during extension builds.
"""

# Extension file suffix mappings
# Convert .cu/.cuh to .mu/.muh so torch_musa's musa_compile rule works correctly
# The musa_compile rule in torch_musa only adds -x musa for .mu/.muh files
# Without this conversion, .cu files would be treated as CUDA files by mcc
# and the --offload-arch=mp_XX flag would fail with clang's CUDA support
EXT_REPLACED_MAPPING = {
    "cuh": "muh",
    "cu": "mu",
    "cc": "cc",
    "cpp": "cpp",
    "cxx": "cxx",
}

# Comprehensive CUDA to MUSA symbol mapping
_MAPPING_RULE = {
    # ATen header mappings (use torch_musa specific paths)
    "#include <ATen/cuda/Atomic.cuh>": '#include "torch_musa/share/generated_cuda_compatible/include/ATen/musa/Atomic.muh"',
    "#include <ATen/cuda/CUDAContext.h>": '#include "torch_musa/csrc/aten/musa/MUSAContext.h"',
    "#include <ATen/cuda/CUDADataType.h>": '#include "torch_musa/csrc/aten/musa/MUSADtype.muh"',
    "#include <ATen/cuda/CUDAGeneratorImpl.h>": '#include "torch_musa/csrc/aten/musa/CUDAGeneratorImpl.h"',
    "#include <ATen/cuda/detail/UnpackRaw.cuh>": '#include "torch_musa/csrc/aten/musa/UnpackRaw.muh"',
    "#include <ATen/cuda/Exceptions.h>": '#include "torch_musa/csrc/aten/musa/Exceptions.h"',
    "at::cuda": "at::musa",
    "torch::kCUDA": "torch::kMUSA",
    # File extension mappings for include statements (.cuh -> .muh)
    # Both quoted includes and angle-bracket includes
    '.cuh"': '.muh"',
    ".cuh>": ".muh>",
    # C10 header mappings (use torch_musa specific paths)
    "#include <c10/cuda/CUDAException.h>": '#include "torch_musa/csrc/core/MUSAException.h"',
    "#include <c10/cuda/CUDAGuard.h>": '#include "torch_musa/csrc/core/MUSAGuard.h"',
    "#include <c10/cuda/CUDAStream.h>": '#include "torch_musa/csrc/core/MUSAStream.h"',
    "c10::cuda": "c10::musa",
    # C10 CUDA macros
    "C10_CUDA_KERNEL_LAUNCH_CHECK": "C10_MUSA_KERNEL_LAUNCH_CHECK",
    "C10_CUDA_CHECK": "C10_MUSA_CHECK",
    "C10_CUDA_ERROR_HANDLED": "C10_MUSA_ERROR_HANDLED",
    "C10_CUDA_IGNORE_ERROR": "C10_MUSA_IGNORE_ERROR",
    # C10 CUDA header file mappings (must come before generic c10/cuda mapping)
    "c10/cuda/CUDAException.h": "c10/musa/MUSAException.h",
    "c10/cuda/CUDAStream.h": "c10/musa/MUSAStream.h",
    "c10/cuda/CUDAGuard.h": "c10/musa/MUSAGuard.h",
    "c10/cuda/CUDAFunctions.h": "c10/musa/MUSAFunctions.h",
    "c10/cuda/CUDAMacros.h": "c10/musa/MUSAMacros.h",
    "c10/cuda/CUDACachingAllocator.h": "c10/musa/MUSACachingAllocator.h",
    "<c10/cuda/CUDAStream.h>": '"torch_musa/csrc/core/MUSAStream.h"',
    # Generic c10/cuda mapping (after specific mappings)
    "c10/cuda": "c10/musa",
    # CUDA standard library -> MUSA standard library
    "cuda/std": "musa/std",
    "<cuda/functional>": "<musa/functional>",
    "<cuda/std/": "<musa/std/",
    "#include <cuda/": "#include <musa/",
    # CUDA namespaces and paths
    "torch::cuda": "torch::musa",
    "torch.cuda": "torch.musa",
    "at::kCUDA": "at::kMUSA",
    "at::DeviceType::CUDA": "at::DeviceType::MUSA",
    "c10::DeviceType::CUDA": "c10::DeviceType::MUSA",
    # cuBLAS -> muBLAS
    "cublas": "mublas",
    "CUBLAS": "MUBLAS",
    "cublasHandle_t": "mublasHandle_t",
    "cublasStatus_t": "mublasStatus_t",
    "cublasCreate": "mublasCreate",
    "cublasDestroy": "mublasDestroy",
    "cublasSetStream": "mublasSetStream",
    "cublasGetStream": "mublasGetStream",
    "cublasSgemm": "mublasSgemm",
    "cublasDgemm": "mublasDgemm",
    "cublasHgemm": "mublasHgemm",
    "cublasGemmEx": "mublasGemmEx",
    "cublasGemmBatchedEx": "mublasGemmBatchedEx",
    "cublasGemmStridedBatchedEx": "mublasGemmStridedBatchedEx",
    "cublasSgemmBatched": "mublasSgemmBatched",
    "cublasDgemmBatched": "mublasDgemmBatched",
    "cublasSgemmStridedBatched": "mublasSgemmStridedBatched",
    "cublasDgemmStridedBatched": "mublasDgemmStridedBatched",
    # cuBLASLt specific mappings
    "CUBLASLT_MATMUL_DESC_A_SCALE_POINTER": "MUBLASLT_MATMUL_DESC_A_SCALE_POINTER",
    "CUBLASLT_MATMUL_DESC_B_SCALE_POINTER": "MUBLASLT_MATMUL_DESC_B_SCALE_POINTER",
    "CUBLASLT_MATMUL_DESC_FAST_ACCUM": "MUBLASLT_MATMUL_DESC_FAST_ACCUM",
    "CUBLASLT_MATMUL_DESC_TRANSA": "MUBLASLT_MATMUL_DESC_TRANSA",
    "CUBLASLT_MATMUL_DESC_TRANSB": "MUBLASLT_MATMUL_DESC_TRANSB",
    "CUBLASLT_MATMUL_PREF_MAX_WORKSPACE_BYTES": "MUBLASLT_MATMUL_PREF_MAX_WORKSPACE_BYTES",
    "CUBLASLT_MATRIX_LAYOUT_BATCH_COUNT": "MUBLASLT_MATRIX_LAYOUT_BATCH_COUNT",
    "CUBLASLT_MATRIX_LAYOUT_STRIDED_BATCH_OFFSET": "MUBLASLT_MATRIX_LAYOUT_STRIDED_BATCH_OFFSET",
    "CUBLAS_COMPUTE_32F": "MUBLAS_COMPUTE_32F",
    "CUBLAS_OP_N": "MUBLAS_OP_N",
    "CUBLAS_OP_T": "MUBLAS_OP_T",
    "CUBLAS_STATUS_NOT_SUPPORTED": "MUBLAS_STATUS_NOT_IMPLEMENTED",
    "CUBLAS_STATUS_SUCCESS": "MUBLAS_STATUS_SUCCESS",
    "cublasComputeType_t": "mublasComputeType_t",
    "cublasGetStatusString": "mublasGetStatusString",
    "cublasLtHandle_t": "mublasLtHandle_t",
    "cublasLtMatmul": "mublasLtMatmul",
    "cublasLtMatmulAlgoGetHeuristic": "mublasLtMatmulAlgoGetHeuristic",
    "cublasLtMatmulDescAttributes_t": "mublasLtMatmulDescAttributes_t",
    "cublasLtMatmulDescCreate": "mublasLtMatmulDescCreate",
    "cublasLtMatmulDescDestroy": "mublasLtMatmulDescDestroy",
    "cublasLtMatmulDescOpaque_t": "mublasLtMatmulDescOpaque_t",
    "cublasLtMatmulDescSetAttribute": "mublasLtMatmulDescSetAttribute",
    "cublasLtMatmulDesc_t": "mublasLtMatmulDesc_t",
    "cublasLtMatmulHeuristicResult_t": "mublasLtMatmulHeuristicResult_t",
    "cublasLtMatmulPreferenceAttributes_t": "mublasLtMatmulPreferenceAttributes_t",
    "cublasLtMatmulPreferenceCreate": "mublasLtMatmulPreferenceCreate",
    "cublasLtMatmulPreferenceDestroy": "mublasLtMatmulPreferenceDestroy",
    "cublasLtMatmulPreferenceOpaque_t": "mublasLtMatmulPreferenceOpaque_t",
    "cublasLtMatmulPreferenceSetAttribute": "mublasLtMatmulPreferenceSetAttribute",
    "cublasLtMatmulPreference_t": "mublasLtMatmulPreference_t",
    "cublasLtMatrixLayoutAttribute_t": "mublasLtMatrixLayoutAttribute_t",
    "cublasLtMatrixLayoutCreate": "mublasLtMatrixLayoutCreate",
    "cublasLtMatrixLayoutDestroy": "mublasLtMatrixLayoutDestroy",
    "cublasLtMatrixLayoutOpaque_t": "mublasLtMatrixLayoutOpaque_t",
    "cublasLtMatrixLayoutSetAttribute": "mublasLtMatrixLayoutSetAttribute",
    "cublasLtMatrixLayout_t": "mublasLtMatrixLayout_t",
    # cuRAND -> muRAND
    "curand": "murand",
    "CURAND": "MURAND",
    "curandState": "murandState",
    "curandStatePhilox4_32_10_t": "murandStatePhilox4_32_10_t",
    "curand_init": "murand_init",
    "curand_uniform": "murand_uniform",
    "curand_uniform4": "murand_uniform4",
    "curand_normal": "murand_normal",
    "curand_normal4": "murand_normal4",
    # cuDNN -> muDNN
    "cudnn": "mudnn",
    "CUDNN": "MUDNN",
    "cudnnHandle_t": "mudnnHandle_t",
    "cudnnCreate": "mudnnCreate",
    "cudnnDestroy": "mudnnDestroy",
    # CUDA Runtime
    "cudaMalloc": "musaMalloc",
    "cudaFree": "musaFree",
    "cudaMemcpy": "musaMemcpy",
    "cudaMemcpyAsync": "musaMemcpyAsync",
    "cudaMemset": "musaMemset",
    "cudaMemsetAsync": "musaMemsetAsync",
    "cudaDeviceSynchronize": "musaDeviceSynchronize",
    "cudaStreamSynchronize": "musaStreamSynchronize",
    "cudaGetDevice": "musaGetDevice",
    "cudaSetDevice": "musaSetDevice",
    "cudaGetDeviceCount": "musaGetDeviceCount",
    "cudaGetDeviceProperties": "musaGetDeviceProperties",
    "cudaDeviceGetAttribute": "musaDeviceGetAttribute",
    # CUDA Stream/Event
    "cudaStream_t": "musaStream_t",
    "cudaEvent_t": "musaEvent_t",
    "cudaStreamCreate": "musaStreamCreate",
    "cudaStreamDestroy": "musaStreamDestroy",
    "cudaEventCreate": "musaEventCreate",
    "cudaEventDestroy": "musaEventDestroy",
    "cudaEventRecord": "musaEventRecord",
    "cudaEventSynchronize": "musaEventSynchronize",
    "cudaEventElapsedTime": "musaEventElapsedTime",
    "cudaStreamWaitEvent": "musaStreamWaitEvent",
    # CUDA Error handling
    "cudaError_t": "musaError_t",
    "cudaSuccess": "musaSuccess",
    "cudaGetLastError": "musaGetLastError",
    "cudaGetErrorString": "musaGetErrorString",
    "cudaPeekAtLastError": "musaPeekAtLastError",
    # CUDA Memory types
    "cudaMemcpyHostToDevice": "musaMemcpyHostToDevice",
    "cudaMemcpyDeviceToHost": "musaMemcpyDeviceToHost",
    "cudaMemcpyDeviceToDevice": "musaMemcpyDeviceToDevice",
    "cudaMemcpyHostToHost": "musaMemcpyHostToHost",
    # Data types
    # Note: __half and half are the same in MUSA, no mapping needed
    "__nv_bfloat16": "__mt_bfloat16",
    "__nv_bfloat162": "__mt_bfloat162",
    "nv_bfloat16": "__mt_bfloat16",
    "nv_bfloat162": "__mt_bfloat162",
    "__nv_half": "__half",
    "nv_half": "__half",
    # FP8 data types
    "__NV_E4M3": "__MT_E4M3",
    "__NV_E5M2": "__MT_E5M2",
    "__NV_SATFINITE": "__MT_SATFINITE",
    "__nv_cvt_float2_to_fp8x2": "__musa_cvt_float2_to_fp8x2",
    "__nv_fp8_e4m3": "__mt_fp8_e4m3",
    "__nv_fp8_e5m2": "__mt_fp8_e5m2",
    "__nv_fp8x2_e4m3": "__mt_fp8x2_e4m3",
    "__nv_fp8x2_e5m2": "__mt_fp8x2_e5m2",
    "__nv_fp8x2_storage_t": "__mt_fp8x2_storage_t",
    "__nv_fp8x4_e4m3": "__mt_fp8x4_e4m3",
    "__nv_fp8x4_e5m2": "__mt_fp8x4_e5m2",
    "__nv_fp8x4_storage_t": "__mt_fp8x4_storage_t",
    "#include <cuda_fp8.h>": "#include <musa_fp8.h>",
    "CUDA_R_8F_E4M3": "MUSA_R_8F_E4M3",
    "CUDA_R_8F_E5M2": "MUSA_R_8F_E5M2",
    # Cutlass -> Mutlass
    '#include "cutlass/array.h"': "#include <mutlass/array.h>",
    "#include <cutlass/array.h>": "#include <mutlass/array.h>",
    "#include <cutlass/cutlass.h>": "#include <mutlass/mutlass.h>",
    "#include <cutlass/numeric_types.h>": "#include <mutlass/numeric_types.h>",
    "cutlass::AlignedArray": "mutlass::AlignedArray",
    "cutlass::bfloat16_t": "mutlass::bfloat16_t",
    "cutlass::half_t": "mutlass::half_t",
    "cutlass": "mutlass",
    "CUTLASS": "MUTLASS",
    "cutlass/": "mutlass/",
    "cutlass::": "mutlass::",
    # CUB - MUSA provides cub directly, no conversion needed (no mapping needed)
    # Thrust
    "thrust::cuda": "thrust::musa",
    # NCCL -> MCCL
    "nccl": "mccl",
    "NCCL": "MCCL",
    "ncclComm_t": "mcclComm_t",
    "ncclDataType_t": "mcclDataType_t",
    "ncclRedOp_t": "mcclRedOp_t",
    "ncclResult_t": "mcclResult_t",
    "ncclSuccess": "mcclSuccess",
    # cuSPARSE -> muSPARSE
    "cusparse": "musparse",
    "CUSPARSE": "MUSPARSE",
    "cusparseHandle_t": "musparseHandle_t",
    "cusparseCreate": "musparseCreate",
    "cusparseDestroy": "musparseDestroy",
    # cuSOLVER -> muSOLVER
    "cusolver": "musolver",
    "CUSOLVER": "MUSOLVER",
    "cusolverDnHandle_t": "musolverDnHandle_t",
    "cusolverDnCreate": "musolverDnCreate",
    "cusolverDnDestroy": "musolverDnDestroy",
    # cuFFT -> muFFT
    "cufft": "mufft",
    "CUFFT": "MUFFT",
    "cufftHandle": "mufftHandle",
    "cufftPlan1d": "mufftPlan1d",
    "cufftPlan2d": "mufftPlan2d",
    "cufftPlan3d": "mufftPlan3d",
    "cufftExecC2C": "mufftExecC2C",
    "cufftExecR2C": "mufftExecR2C",
    "cufftExecC2R": "mufftExecC2R",
    # CUDA kernel launch syntax - same in MUSA, no mapping needed
    # CUDA device attributes
    "cudaDevAttrMaxThreadsPerBlock": "musaDevAttrMaxThreadsPerBlock",
    "cudaDevAttrMaxBlockDimX": "musaDevAttrMaxBlockDimX",
    "cudaDevAttrMaxBlockDimY": "musaDevAttrMaxBlockDimY",
    "cudaDevAttrMaxBlockDimZ": "musaDevAttrMaxBlockDimZ",
    "cudaDevAttrMaxGridDimX": "musaDevAttrMaxGridDimX",
    "cudaDevAttrMaxGridDimY": "musaDevAttrMaxGridDimY",
    "cudaDevAttrMaxGridDimZ": "musaDevAttrMaxGridDimZ",
    "cudaDevAttrMaxSharedMemoryPerBlock": "musaDevAttrMaxSharedMemoryPerBlock",
    "cudaDevAttrWarpSize": "musaDevAttrWarpSize",
    "cudaDevAttrMultiProcessorCount": "musaDevAttrMultiProcessorCount",
    # Additional PyTorch CUDA utilities
    "getCurrentCUDAStream": "getCurrentMUSAStream",
    "getDefaultCUDAStream": "getDefaultMUSAStream",
    "CUDAStream": "MUSAStream",
    "CUDAGuard": "MUSAGuard",
    "OptionalCUDAGuard": "OptionalMUSAGuard",
    "CUDAStreamGuard": "MUSAStreamGuard",
    "CUDAEvent": "MUSAEvent",
    # CUDA includes to MUSA
    "cuda_runtime.h": "musa_runtime.h",
    "cuda_runtime_api.h": "musa_runtime_api.h",
    "cuda.h": "musa.h",
    "cuda_fp16.h": "musa_fp16.h",
    "cuda_bf16.h": "musa_bf16.h",
    # Additional CUDA runtime functions
    "cudaHostAlloc": "musaHostAlloc",
    "cudaHostFree": "musaHostFree",
    "cudaMallocHost": "musaMallocHost",
    "cudaFreeHost": "musaFreeHost",
    "cudaMallocManaged": "musaMallocManaged",
    "cudaMallocAsync": "musaMallocAsync",
    "cudaFreeAsync": "musaFreeAsync",
    "cudaMemcpy2D": "musaMemcpy2D",
    "cudaMemcpy2DAsync": "musaMemcpy2DAsync",
    "cudaMemcpy3D": "musaMemcpy3D",
    "cudaMemcpy3DAsync": "musaMemcpy3DAsync",
    "cudaMemGetInfo": "musaMemGetInfo",
    "cudaMemPrefetchAsync": "musaMemPrefetchAsync",
    "cudaPointerGetAttributes": "musaPointerGetAttributes",
    # CUDA stream flags and types
    "cudaStreamDefault": "musaStreamDefault",
    "cudaStreamNonBlocking": "musaStreamNonBlocking",
    "cudaStreamCreateWithFlags": "musaStreamCreateWithFlags",
    "cudaStreamCreateWithPriority": "musaStreamCreateWithPriority",
    "cudaStreamQuery": "musaStreamQuery",
    "cudaStreamGetPriority": "musaStreamGetPriority",
    "cudaStreamGetFlags": "musaStreamGetFlags",
    # CUDA event flags
    "cudaEventDefault": "musaEventDefault",
    "cudaEventBlockingSync": "musaEventBlockingSync",
    "cudaEventDisableTiming": "musaEventDisableTiming",
    "cudaEventCreateWithFlags": "musaEventCreateWithFlags",
    "cudaEventQuery": "musaEventQuery",
    # CUDA memory flags
    "cudaHostAllocDefault": "musaHostAllocDefault",
    "cudaHostAllocPortable": "musaHostAllocPortable",
    "cudaHostAllocMapped": "musaHostAllocMapped",
    "cudaHostAllocWriteCombined": "musaHostAllocWriteCombined",
    "cudaMemoryTypeHost": "musaMemoryTypeHost",
    "cudaMemoryTypeDevice": "musaMemoryTypeDevice",
    "cudaMemoryTypeManaged": "musaMemoryTypeManaged",
    # Device management
    "cudaDeviceReset": "musaDeviceReset",
    "cudaDeviceSetCacheConfig": "musaDeviceSetCacheConfig",
    "cudaDeviceGetCacheConfig": "musaDeviceGetCacheConfig",
    "cudaDeviceSetSharedMemConfig": "musaDeviceSetSharedMemConfig",
    "cudaDeviceGetSharedMemConfig": "musaDeviceGetSharedMemConfig",
    "cudaGetDeviceFlags": "musaGetDeviceFlags",
    "cudaSetDeviceFlags": "musaSetDeviceFlags",
    "cudaDeviceCanAccessPeer": "musaDeviceCanAccessPeer",
    "cudaDeviceEnablePeerAccess": "musaDeviceEnablePeerAccess",
    "cudaDeviceDisablePeerAccess": "musaDeviceDisablePeerAccess",
    # CUDA occupancy
    "cudaOccupancyMaxActiveBlocksPerMultiprocessor": "musaOccupancyMaxActiveBlocksPerMultiprocessor",
    "cudaOccupancyMaxPotentialBlockSize": "musaOccupancyMaxPotentialBlockSize",
    # Additional device properties
    "cudaDeviceProp": "musaDeviceProp",
    "cudaFuncAttributes": "musaFuncAttributes",
    "cudaFuncGetAttributes": "musaFuncGetAttributes",
    "cudaFuncSetAttribute": "musaFuncSetAttribute",
    "cudaFuncSetCacheConfig": "musaFuncSetCacheConfig",
    # CUDA texture/surface (if supported)
    "cudaTextureObject_t": "musaTextureObject_t",
    "cudaSurfaceObject_t": "musaSurfaceObject_t",
    "cudaCreateTextureObject": "musaCreateTextureObject",
    "cudaDestroyTextureObject": "musaDestroyTextureObject",
    "cudaCreateSurfaceObject": "musaCreateSurfaceObject",
    "cudaDestroySurfaceObject": "musaDestroySurfaceObject",
    # CUDA cooperative groups - same in MUSA, no mapping needed
    # Additional cuBLAS functions
    "cublasSetMathMode": "mublasSetMathMode",
    "cublasGetMathMode": "mublasGetMathMode",
    "CUBLAS_DEFAULT_MATH": "MUBLAS_DEFAULT_MATH",
    "CUBLAS_TENSOR_OP_MATH": "MUBLAS_TENSOR_OP_MATH",
    "cublasLtCreate": "mublasLtCreate",
    "cublasLtDestroy": "mublasLtDestroy",
    "cublasLtHandle_t": "mublasLtHandle_t",
    "cublasLtMatmul": "mublasLtMatmul",
    # Additional cuDNN functions
    "cudnnStatus_t": "mudnnStatus_t",
    "cudnnSetStream": "mudnnSetStream",
    "cudnnGetStream": "mudnnGetStream",
    "cudnnTensorDescriptor_t": "mudnnTensorDescriptor_t",
    "cudnnFilterDescriptor_t": "mudnnFilterDescriptor_t",
    "cudnnConvolutionDescriptor_t": "mudnnConvolutionDescriptor_t",
    "cudnnPoolingDescriptor_t": "mudnnPoolingDescriptor_t",
    "cudnnActivationDescriptor_t": "mudnnActivationDescriptor_t",
    "cudnnDropoutDescriptor_t": "mudnnDropoutDescriptor_t",
    "cudnnRNNDescriptor_t": "mudnnRNNDescriptor_t",
    "cudnnCreateTensorDescriptor": "mudnnCreateTensorDescriptor",
    "cudnnDestroyTensorDescriptor": "mudnnDestroyTensorDescriptor",
    "cudnnSetTensor4dDescriptor": "mudnnSetTensor4dDescriptor",
    "cudnnSetTensorNdDescriptor": "mudnnSetTensorNdDescriptor",
    # Flash attention and transformer related - same names, no mapping needed
    # Paged attention (vLLM specific) - same names, no mapping needed
    # Additional NCCL functions
    "ncclCommInitRank": "mcclCommInitRank",
    "ncclCommInitAll": "mcclCommInitAll",
    "ncclCommDestroy": "mcclCommDestroy",
    "ncclCommCount": "mcclCommCount",
    "ncclCommCuDevice": "mcclCommCuDevice",
    "ncclCommUserRank": "mcclCommUserRank",
    "ncclAllReduce": "mcclAllReduce",
    "ncclBroadcast": "mcclBroadcast",
    "ncclReduce": "mcclReduce",
    "ncclAllGather": "mcclAllGather",
    "ncclReduceScatter": "mcclReduceScatter",
    "ncclGroupStart": "mcclGroupStart",
    "ncclGroupEnd": "mcclGroupEnd",
    "ncclSend": "mcclSend",
    "ncclRecv": "mcclRecv",
    "ncclGetUniqueId": "mcclGetUniqueId",
    "ncclUniqueId": "mcclUniqueId",
    # CUDA math intrinsics - same in MUSA, no mapping needed
    # (__shfl_sync, __shfl_xor_sync, __shfl_up_sync, __shfl_down_sync,
    #  __ballot_sync, __any_sync, __all_sync, __syncthreads, __syncwarp,
    #  __threadfence, __threadfence_block, __threadfence_system)
    # Atomic operations - same in MUSA, no mapping needed
    # (atomicAdd, atomicSub, atomicExch, atomicMin, atomicMax, atomicInc,
    #  atomicDec, atomicCAS, atomicAnd, atomicOr, atomicXor)
    # CUDA math functions - same in MUSA, no mapping needed
    # (__float2half, __half2float, __float2half_rn, __float22half2_rn,
    #  __half22float2, __hadd, __hsub, __hmul, __hdiv, __hfma,
    #  __hadd2, __hsub2, __hmul2, __hfma2)
    # Common macros
    "CUDA_KERNEL_LOOP": "MUSA_KERNEL_LOOP",
    "CUDA_1D_KERNEL_LOOP": "MUSA_1D_KERNEL_LOOP",
    "CUDA_2D_KERNEL_LOOP": "MUSA_2D_KERNEL_LOOP",
    "CUDA_NUM_THREADS": "MUSA_NUM_THREADS",
    # GET_BLOCKS and DIVUP are common macros that don't need mapping
    # PyTorch C++ API cuda utilities
    "torch::cuda::getCurrentCUDAStream": "torch::musa::getCurrentMUSAStream",
    "torch::cuda::getDefaultCUDAStream": "torch::musa::getDefaultMUSAStream",
    "torch::cuda::getStreamFromPool": "torch::musa::getStreamFromPool",
    # torch::kCUDA maps to c10::DeviceType::PrivateUse1 (MUSA uses PrivateUse1 device type)
    # This is always available without needing extra includes
    "torch::kCUDA": "c10::DeviceType::PrivateUse1",
    # Device index utilities
    "cudaDeviceIndex": "musaDeviceIndex",
    "CUDADeviceIndex": "MUSADeviceIndex",
    # CUDA driver API -> MUSA driver API
    "CUdeviceptr": "MUdeviceptr",
    "CUdevice": "MUdevice",
    "CUcontext": "MUcontext",
    "CUmodule": "MUmodule",
    "CUfunction": "MUfunction",
    "CUstream": "MUstream",
    "CUevent": "MUevent",
    "CUresult": "MUresult",
    "cuPointerGetAttribute": "muPointerGetAttribute",
    "cuMemGetAddressRange": "muMemGetAddressRange",
    "cuCtxGetCurrent": "muCtxGetCurrent",
    "cuCtxSetCurrent": "muCtxSetCurrent",
    "cuDeviceGet": "muDeviceGet",
    "cuDeviceGetCount": "muDeviceGetCount",
    "cuInit": "muInit",
    "CU_POINTER_ATTRIBUTE_RANGE_START_ADDR": "MU_POINTER_ATTRIBUTE_RANGE_START_ADDR",
    "CU_POINTER_ATTRIBUTE_RANGE_SIZE": "MU_POINTER_ATTRIBUTE_RANGE_SIZE",
    "CU_POINTER_ATTRIBUTE_MEMORY_TYPE": "MU_POINTER_ATTRIBUTE_MEMORY_TYPE",
    "CU_POINTER_ATTRIBUTE_DEVICE_POINTER": "MU_POINTER_ATTRIBUTE_DEVICE_POINTER",
    "CU_POINTER_ATTRIBUTE_HOST_POINTER": "MU_POINTER_ATTRIBUTE_HOST_POINTER",
    "CU_POINTER_ATTRIBUTE_CONTEXT": "MU_POINTER_ATTRIBUTE_CONTEXT",
    "CUDA_SUCCESS": "MUSA_SUCCESS",
    "CUDA_ERROR_INVALID_VALUE": "MUSA_ERROR_INVALID_VALUE",
    "CUDA_ERROR_OUT_OF_MEMORY": "MUSA_ERROR_OUT_OF_MEMORY",
    "CUDA_ERROR_NOT_INITIALIZED": "MUSA_ERROR_NOT_INITIALIZED",
    # HIP -> MUSA mappings (for USE_ROCM code paths)
    "<hip/hip_runtime.h>": "<musa_runtime.h>",
    "hip/hip_runtime.h": "musa_runtime.h",
    "<hipcub/hipcub.hpp>": "<cub/cub.cuh>",
    "hipcub/hipcub.hpp": "cub/cub.cuh",
    "hipcub::": "cub::",
    "__hip_bfloat16": "__mt_bfloat16",
    # THC headers
    "#include <THC/THCAtomics.cuh>": "#include <THC/THCAtomics.muh>",
    # Template keyword fixes for mcc compiler
    ".FlagHeads<VEC_SIZE>": ".template FlagHeads<VEC_SIZE>",
    ".InclusiveSum<VEC_SIZE>": ".template InclusiveSum<VEC_SIZE>",
    ".Reduce<VEC_SIZE>": ".template Reduce<VEC_SIZE>",
    ".Sum<VEC_SIZE>": ".template Sum<VEC_SIZE>",
    "::cast<vec_size>": "::template cast<vec_size>",
    "SCHEDULER::execute": "SCHEDULER::template execute",
    # CUDA launch attributes
    "cudaLaunchAttribute": "musaLaunchAttribute",
    "cudaLaunchAttributeProgrammaticStreamSerialization": "musaLaunchAttributeIgnore",
    "cudaLaunchConfig_t": "musaLaunchConfig_t",
    # FlashInfer specific mappings
    ".is_cuda()": ".is_privateuseone()",
    "->philox_cuda_state": "->philox_musa_state",
    # CUDA arch guards
    "(__CUDA_ARCH__ < 800)": "(__MUSA_ARCH__ < 220)",
    "(__CUDA_ARCH__ >= 900)": "(__MUSA_ARCH__ >= 310)",
    # cuda::std namespace
    "cuda::std::numeric_limits": "musa::std::numeric_limits",
    "#include <cuda/std/functional>": "#include <musa/std/functional>",
    "#include <cuda/std/limits>": "#include <musa/std/limits>",
    # Compute capability checks
    "compute_capacity.first >= 8": "compute_capacity.first >= 3",
    # FlashInfer vec_dtypes.cuh - mcc compiler requires 'template' keyword
    "::cast<vec_size>": "::template cast<vec_size>",
    # FlashInfer norm.cuh - replace math.cuh with MUSA fast math functions
    '#include "math.cuh"': """
// MUSA fast math intrinsics (replacing flashinfer::math functions)
__device__ __forceinline__ float fast_rsqrtf(float x) { return __frsqrt_rn(x); }
__device__ __forceinline__ float fast_rcp(float x) { return __frcp_rn(x); }
""",
    # FlashInfer norm.cuh - replace math:: functions with MUSA intrinsics
    "math::shfl_xor_sync(sum_sq, offset);": "__shfl_xor_sync(0xffffffff, sum_sq, offset);",
    "math::rsqrt(smem[0] / float(d) + eps);": "fast_rsqrtf(smem[0] / float(d) + eps);",
    # FlashInfer sampling.cuh - replace math::ptx_rcp with fast_rcp
    "math::ptx_rcp(max(sum_low, 1e-8));": "fast_rcp(max(sum_low, 1e-8));",
    "math::ptx_rcp(denom);": "fast_rcp(denom);",
    # PTX assembly -> MUSA compatible code
    # ld.global.nc.b64 is non-cached load, st.global.cg.b64 is cache-global store
    # Replace with regular load/store since MUSA doesn't support PTX
    'asm volatile("ld.global.nc.b64 %0,[%1];" : "=l"(tmp) : "l"(src + j) : "memory");': "tmp = src[j];",
    'asm volatile("st.global.cg.b64 [%0],%1;" ::"l"(dst + j), "l"(tmp) : "memory");': "dst[j] = tmp;",
    # Memory barrier and atomic operations - replace PTX with atomics and fence
    'asm volatile("st.release.sys.global.u32 [%1], %0;" ::"r"(flag), "l"(flag_addr));': "__threadfence_system(); atomicExch(flag_addr, flag);",
    'asm volatile("membar.sys; st.volatile.global.u32 [%1], %0;" ::"r"(flag), "l"(flag_addr));': "__threadfence_system(); atomicExch(flag_addr, flag);",
    'asm volatile("ld.acquire.sys.global.u32 %0, [%1];" : "=r"(flag) : "l"(flag_addr));': "flag = atomicAdd(flag_addr, 0); __threadfence_system();",
    'asm volatile("ld.volatile.global.u32 %0, [%1]; membar.gl;" : "=r"(flag) : "l"(flag_addr));': "flag = atomicAdd(flag_addr, 0); __threadfence();",
    'asm volatile("st.volatile.global.u32 [%1], %0;" ::"r"(flag), "l"(flag_addr));': "atomicExch(flag_addr, flag);",
    'asm volatile("ld.volatile.global.u32 %0, [%1];" : "=r"(flag) : "l"(flag_addr));': "flag = atomicAdd(flag_addr, 0);",
    # MUSA compiler issue: __restrict__ in struct members causes copy issues
    # Remove __restrict__ from RankData.ptrs array
    "const void* __restrict__ ptrs[8]": "const void* ptrs[8]",
}



"""
torchada.cuda - CUDA-compatible API that works on both CUDA and MUSA platforms.

This module provides the same interface as torch.cuda but automatically
routes to torch.musa on Moore Threads hardware.

Note: After importing torchada, you can use standard torch.cuda APIs directly.
This module is provided for internal use and backwards compatibility.

Usage (preferred):
    import torchada  # Apply patches
    import torch

    if torch.cuda.is_available():
        torch.cuda.set_device(0)
        tensor = tensor.cuda()
"""

from typing import Optional, Union

from .._platform import Platform, detect_platform


def _get_backend():
    """Get the appropriate backend module (torch.cuda or torch.musa)."""
    platform = detect_platform()

    if platform == Platform.MUSA:
        import torch
        import torch_musa

        return torch.musa
    elif platform == Platform.CUDA:
        import torch

        return torch.cuda
    else:
        # Return torch.cuda for API compatibility, even if not available
        import torch

        return torch.cuda


# Core device functions
def is_available() -> bool:
    """Check if CUDA/MUSA is available."""
    return _get_backend().is_available()


def device_count() -> int:
    """Return the number of GPUs available."""
    backend = _get_backend()
    if hasattr(backend, "device_count"):
        return backend.device_count()
    return 0


def current_device() -> int:
    """Return the index of the currently selected device."""
    return _get_backend().current_device()


def set_device(device: Union[int, str, "torch.device"]) -> None:
    """Set the current device."""
    _get_backend().set_device(device)


def get_device_name(device: Optional[Union[int, str]] = None) -> str:
    """Get the name of a device."""
    return _get_backend().get_device_name(device)


def get_device_capability(device: Optional[Union[int, str]] = None) -> tuple:
    """Get the CUDA/MUSA compute capability of a device."""
    return _get_backend().get_device_capability(device)


def get_device_properties(device: Optional[Union[int, str]] = None):
    """Get the properties of a device."""
    return _get_backend().get_device_properties(device)


# Memory management
def memory_allocated(device: Optional[Union[int, str]] = None) -> int:
    """Return the current GPU memory occupied by tensors in bytes."""
    return _get_backend().memory_allocated(device)


def max_memory_allocated(device: Optional[Union[int, str]] = None) -> int:
    """Return the maximum GPU memory occupied by tensors in bytes."""
    return _get_backend().max_memory_allocated(device)


def memory_reserved(device: Optional[Union[int, str]] = None) -> int:
    """Return the current GPU memory managed by the caching allocator in bytes."""
    return _get_backend().memory_reserved(device)


def max_memory_reserved(device: Optional[Union[int, str]] = None) -> int:
    """Return the maximum GPU memory managed by the caching allocator in bytes."""
    return _get_backend().max_memory_reserved(device)


def memory_cached(device: Optional[Union[int, str]] = None) -> int:
    """Deprecated: Use memory_reserved instead."""
    return _get_backend().memory_cached(device)


def max_memory_cached(device: Optional[Union[int, str]] = None) -> int:
    """Deprecated: Use max_memory_reserved instead."""
    return _get_backend().max_memory_cached(device)


def empty_cache() -> None:
    """Release all unoccupied cached memory."""
    _get_backend().empty_cache()


def reset_peak_memory_stats(device: Optional[Union[int, str]] = None) -> None:
    """Reset the peak memory stats."""
    _get_backend().reset_peak_memory_stats(device)


def reset_max_memory_allocated(device: Optional[Union[int, str]] = None) -> None:
    """Reset the starting point in tracking maximum GPU memory occupied."""
    _get_backend().reset_max_memory_allocated(device)


def reset_max_memory_cached(device: Optional[Union[int, str]] = None) -> None:
    """Reset the starting point in tracking maximum GPU memory managed."""
    _get_backend().reset_max_memory_cached(device)


# Synchronization
def synchronize(device: Optional[Union[int, str]] = None) -> None:
    """Wait for all kernels in all streams on a device to complete."""
    _get_backend().synchronize(device)


# Stream and Event classes - will be set up dynamically
def _setup_stream_event_classes():
    """Set up Stream and Event classes from the backend."""
    backend = _get_backend()

    # These will be the actual classes from the backend
    global Stream, Event, current_stream, default_stream, stream

    Stream = backend.Stream if hasattr(backend, "Stream") else None
    Event = backend.Event if hasattr(backend, "Event") else None

    if hasattr(backend, "current_stream"):
        current_stream = backend.current_stream
    if hasattr(backend, "default_stream"):
        default_stream = backend.default_stream
    if hasattr(backend, "stream"):
        stream = backend.stream


# Initialize stream/event classes
try:
    _setup_stream_event_classes()
except:
    Stream = None
    Event = None
    current_stream = None
    default_stream = None
    stream = None



"""
NVTX stub module for MUSA platform.

This module provides no-op implementations of NVTX (NVIDIA Tools Extension)
functions for profiling. On MUSA platform, these are stubs that do nothing.

Usage:
    import torchada  # Apply patches first
    import torch.cuda.nvtx as nvtx
    
    nvtx.range_push("my_range")
    # ... code to profile ...
    nvtx.range_pop()
    
    # Or use the context manager
    with nvtx.range("my_range"):
        # ... code to profile ...
"""

from contextlib import contextmanager
from typing import Optional

__all__ = [
    "mark",
    "range",
    "range_push",
    "range_pop",
    "range_start",
    "range_end",
]


def mark(msg: str) -> None:
    """
    Mark an instantaneous event in the timeline.

    This is a no-op on MUSA platform.

    Args:
        msg: The message to associate with the mark.
    """
    pass


def range_push(msg: str) -> int:
    """
    Push a range onto the stack.

    This is a no-op on MUSA platform.

    Args:
        msg: The message to associate with the range.

    Returns:
        The zero-based depth of the range that is started.
    """
    return 0


def range_pop() -> None:
    """
    Pop a range off the stack.

    This is a no-op on MUSA platform.
    """
    pass


def range_start(msg: str) -> int:
    """
    Start a range.

    This is a no-op on MUSA platform.

    Args:
        msg: The message to associate with the range.

    Returns:
        A range ID that can be passed to range_end.
    """
    return 0


def range_end(range_id: int) -> None:
    """
    End a range.

    This is a no-op on MUSA platform.

    Args:
        range_id: The range ID returned by range_start.
    """
    pass


@contextmanager
def range(msg: str, *args, **kwargs):
    """
    Context manager for NVTX ranges.

    This is a no-op on MUSA platform.

    Args:
        msg: The message to associate with the range.

    Yields:
        None
    """
    yield



#!/usr/bin/env python3
"""
Setup script for torchada.

This is a shim for editable installs. The actual configuration is in pyproject.toml.
"""

from setuptools import setup

if __name__ == "__main__":
    setup()

